<?php

class Kohana_Account implements ArrayAccess {
    
    protected static $_instance;
    protected $_session;
    protected $_config;
    protected $_model;
    protected $_userInfo;

    /**
     * Вовзращает экземпляр класса (singleton-паттерн)
     *
     * @return  instance
     */     
    public static function instance() {
        if(!isset(self::$_instance))
        {
            $config = Kohana::$config->load('account');
            self::$_instance = new self($config);
        }
        return self::$_instance;
    }
    
    private function __construct($config = array()) {
        $this->_config = $config;
        $this->_session = Session::instance();
        $this->_model = new Model_Account;
        $this->_config['hash_key'] = $this->_model->getHashKey();
        $this->_config['hash_method'] = 'sha256';
        if($this->isSignedIn())
        {
            $id = $this->_session->get('ID');
            $this->_userInfo = $this->_getInfoFromDB($id);
        }
    }
    
    public function isLoginExists($login)
    {
        $login_count = $this->_model->getAccNumByLogin($login);
        return $login_count > 0;
    }
    
    public function isMailExists($email)
    {
        $email_count = $this->_model->getAccNumByMail($email);
        return $email_count > 0;
    }

    /**
     * Регистрирует нового пользователя и осуществляет вход.
     * Проверяет корректность кода активации и существование аккаунтов с такими же авторизационными данными.
     *
     * @param  string $code Код активации
     * @param  string $email E-Mail адресс
     * @param  string $login
     * @param  string $password
     * @return  array
     */          
    public function signUp($code, $email, $login, $password)
    {
        if($this->_model->isActivationCodeValid($code))
        {
            if(!$this->isMailExists($email) && !$this->isLoginExists($login))
            {
                $id = $this->_model->activateAccount($login, sha1($password), $email, $code);
                $this->completeSignIn($id, $this->hash($password));
                return array(true, 'ok');
            }
            else
            {
                if($this->isLoginExists($login))
                    return array(false, 'login_exists');
                if($this->isMailExists($email))
                    return array(false, 'mail_exists');
            }
        }
        else
        {
            return array(false, 'invalid_code');
        }
    }

    /**
     * Проверяет корректность авторизационных данных и возвращает true, если авторизация прошла успешно, и false, если данные являются некорректными.
     *
     * @param  string $login
     * @param  string $password
     * @return  bool
     */      
    public function signIn($login, $password) {
        $id = $this->_model->checkAuth($login, sha1($password));
        if($id == -1)
            return false;
        else
            return $this->completeSignIn ($id, $this->hash($password));
    }
    
    protected function completeSignIn($id, $passhash) {
        $userHash = $this->hash($id.Request::$user_agent.Request::$client_ip).$this->_config['hash_key'];
        $passhash = $this->hash($passhash.$this->_config['hash_key']);
        Cookie::set('userhash', $passhash);
        $this->_userInfo = $this->_getInfoFromDB($id);
        $this->_session->regenerate();
        $this->_session->set('ID', $id);
        $this->_session->set('LoggedIn', true);
        $this->_session->set('UserHash', $this->hash($userHash));
        $this->_session->set('PasswordHash', $passhash);
        return TRUE;
    }    

    /**
     * Проверяет авторизационный статус пользователя и, если пользователь имеет useragent и IP, отличные от хранимых в сессии, осуществляет выход из текущего сеанса.
     *
     * @return  bool
     */   
    public function isSignedIn()
    {
        if($this->_session->get('LoggedIn'))
        {
            if(!$this->checkHash())
            {
                $this->completeSignOut();
                return false;
            }
        }
        return $this->_session->get('LoggedIn');
    }
    
    protected function checkHash()
    {
        $id = $this->_session->get('ID');
        $userHash = $this->hash($id.Request::$user_agent.Request::$client_ip).$this->_config['hash_key'];
        $userCheck = $this->_session->get('UserHash') == $this->hash($userHash);
        $passCheck = Cookie::get('userhash') == $this->_session->get('PasswordHash');
        return $userCheck AND $passCheck;
    }    
   
    /**
     * Завершает текущий сеанс пользователя.
     *
     * @return  array
     */    
    public function signOut()
    {
        if($this->isSignedIn())
        {
            return $this->completeSignOut();
        }
    }
    
    protected function completeSignOut()
    {
        $this->_session->set('ID', FALSE)->set('LoggedIn', FALSE)->set('UserHash', FALSE);
        Cookie::delete('userhash');
        unset($this->_userInfo);
        $this->_session->restart();
        return TRUE;
    }

    /**
     * Проверяет корректность данного пароля для текущего пользователя.
     *
     * @param string $password Пароль
     * @return  bool
     */        
    public function checkPassword($password) {
        if(!$this->isSignedIn())
            return FALSE;
        $passhash = $this->hash($password);
        return $this->hash($passhash.$this->_config['hash_key']) === $this->_session->get('PasswordHash');
    }
    
    public function changePassword($old, $new)
    {
        /* if(!$this->isSignedIn() || !$this->checkPassword($old))
            return FALSE; */
        $this->_model->changePassword($this->offsetGet('UserID'), sha1($new));
        $passhash = $this->hash($this->hash($new).$this->_config['hash_key']);
        Cookie::set('userhash', $passhash);
        $this->_session->set('PasswordHash', $passhash);
        return TRUE;
    }
    
    public function changeLogin($login)
    {
        if(!$this->isSignedIn() || $this->isLoginExists($login))
            return FALSE;
        $this->_model->changeLogin($this->offsetGet('UserID'), $login);
        return TRUE;
    }    
    
    public function changeMail($email)
    {
        if(!$this->isSignedIn() || $this->isMailExists($email))
            return FALSE;
        $token = md5(time().$this->offsetGet('EMail').$email);
        $this->_session->set('NewMail_Token', $token);
        $this->_session->set('NewMail_Adress', $email);
        return $token;
    }    
    
    public function completeChangeMail($token)
    {
        $email = $this->_session->get('NewMail_Adress');
        if($token == $this->_session->get('NewMail_Token') AND !$this->isMailExists($email))
        {
            $this->_model->changeMail($this->offsetGet('UserID'), $email);
            return true;
        }
        else {
            return false;
        } 
    }        
    
    // ---------------------------- [INFO] -------------------------------------
    
    /**
     * Возвращает массив, содержащий пользовательские данные.
     *
     * @return  array
     */
    public function getInfoAsArray()
    {
        if($this->isSignedIn())
            return $this->_userInfo;
    }
    
    protected function _getInfoFromDB($id)
    {
        $info = $this->_model->getPersonalInfoByID($id)->offsetGet(0);
        $info += $this->_model->getAccountInfoByID($id)->offsetGet(0);
        $info['EMail'] = $info['E-Mail'];
        unset($info['E-Mail']);
        return $info;
    }

    // ---------------------------- [ARRAY] ------------------------------------
    
    public function offsetSet($offset, $value) {
        if(isset($this->_userInfo[$offset]))
        {
            return $this->_userInfo[$offset];
        }
        else
        {
            throw new Kohana_Exception('Invalid key: '.$offset);
        }
    }

    public function offsetGet($offset) {
        if(isset($this->_userInfo[$offset]))
        {
            return $this->_userInfo[$offset];
        }
        else
            return FALSE;
    }
    
    public function offsetUnset($offset) {
        if(isset($this->_userInfo[$offset]))
        {
            unset($this->_userInfo[$offset]);
        }
    }
    
    public function offsetExists($offset) {
        return isset($this->_userInfo[$offset]);
    }
    
    /**
     * Perform a hmac hash, using the configured method.
     *
     * @param   string  $str  string to hash
     * @return  string
     */
    protected function hash($str)
    {
        if ( ! $this->_config['hash_key'])
        {
            $this->_config['hash_key'] = $key = md5(time().Request::$client_ip);
            $this->_model->setHashKey($key);
        }
        return hash_hmac($this->_config['hash_method'], $str, $this->_config['hash_key']);
    }    
}